/*    $Id: Expression.java,v 1.4 1996/03/22 02:30:43 davidliu Exp $
**
**
**************************************************************************
**    Copyright (C) 1996 David Wanqian Liu
**                       davidliu@mti.sgi.com
**                       http://reality.sgi.com/exmployees/davidliu_mti
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**************************************************************************
*/


package expression;



import java.io.*;
import java.util.Stack;
import java.util.Vector;


/**
  The class representing an expression.
  Legal expression symbols include: all double precision numbers;
  arithmetic operators including '+', '-', '*', '/' and '^';
  left bracket '(' and right bracket ')';
  Three variables including 'x' (or 'X'), 'y' (or 'Y') and 'z' (or 'Z');
  And many unary functions such as 'COS', 'SIN', etc.
  */
public class Expression {
  /**
    The root expression node generated by parsing the expression string.
    */
  public ExprNode enode;

  /**
    The symbol queue created for operator-precedence parsing.
    */
  public Vector symQueue;

  /**
    The operator, expression node stack created for operator-precedence parsing.
    */
  public Stack symStack;



  /**
    Construct an empty expression.
    */
  public Expression() {
    enode = null;

    symQueue = new Vector();
    
    symStack = new Stack();

    /* put EOE on the bottom of the symStack */
    symStack.push(new Symbol(Symbol.SYM_TERMINAL, Symbol.TK_EOE));
  }
    

  /**
    Construct an expression by a string representation.
    */
  public Expression(String expr)  throws ExpressionParseException {
    this();
    
  /**
    The expression string buffer used to store the input expression string.
    */
    ExprStringBufferInputStream strBuffer =
      new ExprStringBufferInputStream(expr);
    StreamTokenizer st = new StreamTokenizer(strBuffer);
    
    st.eolIsSignificant(true);
    st.ordinaryChar('(');
    st.ordinaryChar(')');
    st.ordinaryChar('+');
    st.ordinaryChar('-');
    st.ordinaryChar('*');
    st.ordinaryChar('/');
    st.ordinaryChar('^');

    try {
      parse(st);
    } catch (Exception e) {
      String idx = new String();
      for (int i=0; i<strBuffer.pos()-1; i++)
	idx += " ";
      idx += "^";
      throw new ExpressionParseException
	("Parsing error occured (" +e.getMessage()+ "):\n"
	 + "  " + expr + "\n"
	 + "  " + idx);
    }
  }


  /**
    Append a symbol to the end of an expression.
    Return true if the whole expression is reduced into a single expression node,
    or the expression is empty.
    */
  public boolean appendSymbol(Symbol sym) throws ExpressionParseException {
    /* Do until either the expression reduces into a single node,
       or the symbol is shifted into the symbol stack.
       */
    while (true) {

      /* a is the top most terminal symbol */
      Symbol a = (Symbol) symStack.peek();
      if (a.isExprNode()) {
	Symbol tmp = (Symbol) symStack.pop();
	
	a = (Symbol) symStack.peek();
	symStack.push(tmp);
	
	/* cannot have two expression node next to each other without a
	   function in between in a symbol operator stack */
	if (a.isExprNode())
	  throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_NODE);
      }

      /* 1. If only $ is on the symStack and only $ is on the input,
       *    then we get an empty expression
       * 2. If only $ and a ExprNode is on the symStack and only $ is
       *    on the input, the expression is reduced into a single node */
      if (a.isEOE() && sym.isEOE())
      return true;


      /* if a<sym or a=sym then shift sym on the symStack
	 also append it to the symQueue */
      
      if (a.less(sym) || a.equal(sym)) {
	symStack.push(sym);
	symQueue.addElement(sym);

	return false;
      }

      /* else if a>sym then reduce */
      if (a.greater(sym)) {
	ExprNode reduceNode = new ExprNode();
	Symbol outsym, topsym;
	boolean isENode;

	/* outsym holds last popped out terminal symbol
	   topsym holds the uppermost terminal symbol on symStack
	   */
	do {
	  outsym = (Symbol) symStack.pop();
	  
	  if (outsym.isExprNode()) {
	    reduceNode.push(outsym);
	    outsym = (Symbol) symStack.pop();
	  }

	  reduceNode.push(outsym);

	  topsym = (Symbol) symStack.peek();

	  if (topsym.isExprNode()) {
	    topsym = (Symbol) symStack.pop();
	    reduceNode.push(topsym);
	    topsym = (Symbol) symStack.peek();
	  }
	} while(!topsym.less(outsym));

	/* add reduced node onto the top of the symStack */
	symStack.push(new Symbol(Symbol.SYM_EXPRNODE, reduceNode));

	/* continue until either input symbol can be shifted onto stack,
	   or the expression can be reduced into a single expression node. */
	continue;
      }

      /* else we are seeing some unexpected symbol which does not have precedence
	 relationship with current symbol on the stack */
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);
    }
  }
    


  /**
    Parse the input expression string, returns the root expression node
    of the operator-precedence parse tree.
    */
  void parse(StreamTokenizer st)
    throws IOException, ExpressionParseException {

    /* Form the input symbol queue */
    Symbol inputsym;
    boolean finished = false;

    while (!finished) {				   
      inputsym = this.nextSymbol(st);
      finished = appendSymbol(inputsym);
    }

    enode = ((Symbol)symStack.peek()).enode;
  }



  /**
    Evaluate the expression.
    */
  public double eval() throws ExpressionEvalException {
    if (enode == null)
      return 0.0;

    return enode.eval();
  }
      


  /**
    Convert a string to a symbol.
    */
  static public Symbol toSymbol(String s) {
    if ("ENTER/=".equals(s))
      return new Symbol(Symbol.SYM_TERMINAL, Symbol.TK_EOE);

    try {
      StreamTokenizer st = new StreamTokenizer(new StringBufferInputStream(s));
      
      st.eolIsSignificant(true);
      st.ordinaryChar('(');
      st.ordinaryChar(')');
      st.ordinaryChar('+');
      st.ordinaryChar('-');
      st.ordinaryChar('*');
      st.ordinaryChar('/');
      st.ordinaryChar('^');
      
      return nextSymbol(st);
    } catch(Exception e) {
      return new Symbol();
    };
  }


  /**
    Fetch the next symbol from the input expression string.
    */
  static public Symbol nextSymbol(StreamTokenizer st)
    throws IOException, ExpressionParseException {

    int ttype = st.nextToken();

    switch(ttype) {
    default:
      return new Symbol();

    case StreamTokenizer.TT_EOF:
    case StreamTokenizer.TT_EOL:
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_EOE);

    case StreamTokenizer.TT_NUMBER:
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_NUMBER,
			st.nval);

    case StreamTokenizer.TT_WORD:
      if ("A".equals(st.sval.toUpperCase()) ||
	  "B".equals(st.sval.toUpperCase()) ||
	  "C".equals(st.sval.toUpperCase()) ||
	  "D".equals(st.sval.toUpperCase()) ||
	  "F".equals(st.sval.toUpperCase()) ||
	  "G".equals(st.sval.toUpperCase()) ||
	  "H".equals(st.sval.toUpperCase()) ||
	  "I".equals(st.sval.toUpperCase()) ||
	  "J".equals(st.sval.toUpperCase()) ||
	  "K".equals(st.sval.toUpperCase()) ||
	  "L".equals(st.sval.toUpperCase()) ||
	  "M".equals(st.sval.toUpperCase()) ||
	  "N".equals(st.sval.toUpperCase()) ||
	  "O".equals(st.sval.toUpperCase()) ||
	  "P".equals(st.sval.toUpperCase()) ||
	  "Q".equals(st.sval.toUpperCase()) ||
	  "R".equals(st.sval.toUpperCase()) ||
	  "S".equals(st.sval.toUpperCase()) ||
	  "T".equals(st.sval.toUpperCase()) ||
	  "U".equals(st.sval.toUpperCase()) ||
	  "V".equals(st.sval.toUpperCase()) ||
	  "W".equals(st.sval.toUpperCase()) ||
	  "X".equals(st.sval.toUpperCase()) ||
	  "Y".equals(st.sval.toUpperCase()) ||
	  "Z".equals(st.sval.toUpperCase()) ||
	  "E".equals(st.sval) ||
	  "PI".equals(st.sval.toUpperCase()))
	return new Symbol(Symbol.SYM_TERMINAL,
			  Symbol.TK_VARIABLE,
			  st.sval);

      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_FUNCTION,
			st.sval);
      
    case '(':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_BRACKETL);
      
    case ')':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_BRACKETR);
      
    case '+':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_PLUS);
      
    case '-':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_MINUS);

    case '*':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_MULT);

    case '/':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_DIV);

    case '^':
      return new Symbol(Symbol.SYM_TERMINAL,
			Symbol.TK_POW);
    };
  }


  /**
    Return a string indicates where the parsing error occurs.
    */
  public String getErrMsg() {
    String errmsg = new String();
    
    for (int i=0; i<symQueue.size(); i++)
      errmsg += ((Symbol) symQueue.elementAt(i)).toString();

    errmsg += "^";

    return errmsg;
  }

  
}
